#!/usr/bin/env python3
# coding: utf-8

import argparse
from importlib import import_module

from ifconf.printconf import PrintConfigAction

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('config_function'
                        , nargs='+'
                        , type=str
                        , help='configuration callback methos')
    args = parser.parse_args()
    action = PrintConfigAction(parser, args)
    PrintConfigAction.set_callback_methods([getattr(import_module(f.rsplit('.', 1)[0]), f.rsplit('.', 1)[1]) for f in args.config_function])
    action(parser, args, args.config_function)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import argparse
import logging

from ifconf.common import config_callback
from ifconf.config import Config
from ifconf.loader import ConfigLoader
from ifconf.loggingconf import configure_logging
from ifconf.printconf import PrintConfigAction


__MAIN_CONFIG__ = None

def get_main_config():
    if __MAIN_CONFIG__ is None:
        raise RuntimeError('Main configuration must be done before calling configure_module.')
    return __MAIN_CONFIG__

def clear_main_cofig():
    global __MAIN_CONFIG__
    __MAIN_CONFIG__ = None


def configure_main(argparser = None
                   , with_default_args = True
                   , config_arg = 'config.ini'
                   , config_path = []
                   , with_config_logging = True
                   , callback_methods = []):
    global __MAIN_CONFIG__
    __MAIN_CONFIG__ = Config(argparser)

    callback_methods = callback_methods if hasattr(callback_methods, '__iter__') else [callback_methods]
    callback_methods = [config_callback()(m) for m in callback_methods]
    
    if with_default_args:
        add_default_argument(__MAIN_CONFIG__.argparser, config_arg)
        PrintConfigAction.set_callback_methods(callback_methods)
        
    __MAIN_CONFIG__.parse(config_path)
    
    if with_config_logging:
        configure_logging(__MAIN_CONFIG__)
    else:
        __MAIN_CONFIG__.logger = logging.getLogger()
        
    for m in callback_methods:
        loader = ConfigLoader.load(m, __MAIN_CONFIG__)
        try:
            loader.configure(True)
        except Exception as e:
            __MAIN_CONFIG__.err.append('モジュール[{}]の設定取得に失敗しました。エラー：{}'.format(loader.section, e))
    for e in __MAIN_CONFIG__.err:
        __MAIN_CONFIG__.logger.warning(e)
    __MAIN_CONFIG__.logger.info('設定が完了しました。設定ファイル：{}'.format(__MAIN_CONFIG__.config_path))
    return __MAIN_CONFIG__

def configure_main_custom(argparser = None):
    global __MAIN_CONFIG__
    __MAIN_CONFIG__ = Config(argparser)
    return __MAIN_CONFIG__

def add_default_argument(argparser, config_path = None):
    if config_path is not None:
        argparser.add_argument('-c', '--config'
                               , metavar='PATH'
                               , default=config_path
                               , help='設定ファイルへのパス')
    argparser.add_argument('--verbose', '-v'
                           , action='count'
                           , default=False
                           , help='詳細なデータを出力')
    argparser.add_argument('--current_dir'
                           , default='.'
                           , help='カレントディレクトリを指定して実行')
    argparser.add_argument('--debug'
                           , action="store_true"
                           , default=False
                           , help='デバグモードで実行')
    argparser.add_argument('--debug_file'
                           , metavar='FILE'
                           , help='デバグモードで実行しファイル出力')
    argparser.add_argument('--print_conf'
                           , action=PrintConfigAction
                           , default=False
                           , help='設定ファイルの内容を出力')
    return argparser

    




# -*- coding: utf-8 -*-

from functools import reduce

from ifconf.common import config_callback
from ifconf.main import get_main_config
from ifconf.loader import ConfigLoader

def configure_module(*callback_methods, override = {}, immutable = True):
    config = get_main_config()
    loaders = [ConfigLoader.load(config_callback()(callback), config) for callback in callback_methods]
    loaders[0].prepend_name_value_dict(override)
    return reduce(lambda a,b: a.append_name_values(b), loaders).configure(immutable)




# -*- coding: utf-8 -*-

import os
import abc

def get_module_name_for(obj):
    if obj.__module__  == '__main__':
        try:
            return os.path.splitext(os.path.basename(sys.modules['__main__'].__file__))[0]
        except:
            return os.path.splitext(os.path.basename(sys.executable))[0]
    else:
        return obj.__module__
    
def config_callback(section = None):
    def _decorator(func):
        if hasattr(func, '__SECTION__'):
            return func
        func.__MODULE_NAME__ = get_module_name_for(func)
        func.__SECTION__ = section if type(section) == str and section else '{}_{}'.format(func.__MODULE_NAME__, func.__name__)
        return func
    return _decorator(section) if callable(section) else _decorator


class ConfigBuilder(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def add_attr(self, name, default=None, required=False, hidden=False, help=''):
        pass

    @abc.abstractmethod
    def add_attr_boolean(self, name, default=False, required=False, hidden=False, help=''):
        pass

    @abc.abstractmethod
    def add_attr_int(self, name, default=0, required=False, hidden=False, help=''):
        pass

    @abc.abstractmethod
    def add_attr_float(self, name, default=0.0, required=False, hidden=False, help=''):
        pass

    @abc.abstractmethod
    def add_attr_dict(self, name, default={}, required=False, hidden=False, help=''):
        pass
        
    @abc.abstractmethod
    def add_attr_list(self, name, default=[], required=False, hidden=False, help=''):
        pass
        
    @abc.abstractmethod
    def add_attr_path(self, name, default=None, required=False, hidden=False, help=''):
        pass
    



# -*- coding: utf-8 -*-

import os
import os.path

import json

from pathlib import Path

import argparse
import configparser

class Config:

    def __init__(self, argparser = None):
        self.argparser = argparser if argparser is not None else argparse.ArgumentParser()
        self.parser = configparser.ConfigParser()
        self.err = []
        
    def parse(self, config_path = None):
        self.args = self.argparser.parse_args()
        self.config_path = []
        
        if hasattr(self.args, 'config') and self.args.config:
            for p in self.args.config.split(os.pathsep):
                self.config_path.append(p)

        if hasattr(self.args, 'current_dir') and self.args.current_dir:
            if self.args.current_dir != '.':
                os.chdir(self.args.current_dir)

        if config_path:
            if type(config_path) != str and hasattr(config_path, '__iter__'):
                for p in config_path:
                    self.config_path.append(p)
            else:
                self.config_path.append(config_path)
                
        for p in reversed(self.config_path):
            if os.path.exists(p):
                try:
                    with open(p, "r") as f:
                        self.parser.read_file(f)
                except Exception as e:
                    self.err.append('設定ファイル[{}]の読み込みに失敗しました。エラー：{}'.format(p, e))
            else:
                self.err.append('設定ファイル[{}]が見つかりません。'.format(p))

    def get_attr(self, section, option, default='', required=False):
        if required or self.parser.has_option(section, option):
            return self.parser.get(section, option)
        else:
            return default

    def get_attr_boolean(self, section, option, default=False, required=False):
        if required or self.parser.has_option(section, option):
            return self.parser.getboolean(section, option)
        else:
            return default
        
    def get_attr_int(self, section, option, default=0, required=False):
        if required or self.parser.has_option(section, option):
            return self.parser.getint(section, option)
        else:
            return default

    def get_attr_float(self, section, option, default=0.0, required=False):
        if required or self.parser.has_option(section, option):
            return self.parser.getfloat(section, option)
        else:
            return default

    def get_attr_dict(self, section, option, default={}, required=False):
        return self.get_attr_json(section, option, default, required)
        
    def get_attr_list(self, section, option, default=[], required=False):
        return self.get_attr_json(section, option, default, required)
        
    def get_attr_json(self, section, option, default, required):
            if required or self.parser.has_option(section, option):
                value_string = self.parser.get(section, option)
                try:
                    return json.loads(value_string)
                except Exception as e:
                    raise ValueError('Invalid Data Format [{}]. [{}][{}][{}]'.format(e, section, option, value_string), e)
            else:
                return default

    def get_attr_path(self, section, option, default, required):
        if required or self.parser.has_option(section, option):
            return Path(self.parser.get(section, option))
        else:
            return default if type(default) == Path else Path(default)
        



# -*- coding: utf-8 -*-

from collections import namedtuple

from ifconf.config import Config
from ifconf.common import ConfigBuilder

class ConfigLoader(ConfigBuilder):
    
    @classmethod
    def load(cls, callback_method, config):
        assert callback_method is not None, 'callback_method cannot be null.'
        assert hasattr(callback_method, '__MODULE_NAME__'), 'MODULE_NAME must be set for callback_method.'
        assert hasattr(callback_method, '__SECTION__'), 'SECTION must be set for callback_method.'
        loader = ConfigLoader(callback_method.__SECTION__, callback_method.__MODULE_NAME__, config)
        callback_method(loader)
        return loader

    def __init__(self, section, module_name, config):
        assert section is not None, 'section cannot be null.'
        assert module_name is not None, 'module_name cannot be null.'
        assert config is not None, 'config cannot be null.'
        assert type(config) is Config, 'invalid config type:[{}]'.format(type(config))
        self.section = section
        self.module_name = module_name
        self.main_config = config
        self.names = []
        self.values = []

    def append_name_values(self, loader):
        assert loader is not None, 'configloader ancestor cannot be null.'
        for name, value in zip(loader.names, loader.values):
            while name in self.names:
                name = name + '_'
            self.names.append(name)
            self.vlaues.append(vlaue)
        return self

    def prepend_name_value_dict(self, override_dict):
        if override_dict:
            if not isinstance(override_dict, dict):
                raise ValueError('override_dict was expected an instance of dict but was [{}]'.format(type(override_dict)))
            for key, value in override_dict.items():
                name = key
                while name in self.names:
                    name = name + '_'
                if name != key:
                    index = self.names.index(key)
                    self.names.pop(index)
                    self.names.append(name)
                    self.values.append(self.values.pop(index))
                def get_value(conf):
                    return value
                self.names.append(key)
                self.values.append(get_value)
        return self

    def configure(self, immutable):
        assert self.main_config is not None, 'main_config cannot be null.'
        classname = self.section.replace('.','_')
        ntp = namedtuple(classname, self.names) if immutable else self.load_mutble_namedtuple(classname, self.names)
        args = [f(self.main_config) for f in self.values]
        conf = ntp(*args)
        self.main_config.logger.debug(conf)
        return conf

    def load_mutble_namedtuple(self, classname, names):
        from recordclass import recordclass
        return recordclass(classname, names)
        
    def add_attr(self, name, default=None, required=False, hidden=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr(self.section, name, default, required))

    def add_attr_boolean(self, name, default=False, required=False, hidden=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_boolean(self.section, name, default, required))

    def add_attr_int(self, name, default=0, required=False, hidden=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_int(self.section, name, default, required))

    def add_attr_float(self, name, default=0.0, required=False, hidden=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_float(self.section, name, default, required))

    def add_attr_dict(self, name, default={}, required=False, hidden=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_dict(self.section, name, default, required))
        
    def add_attr_list(self, name, default=[], required=False, hidden=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_list(self.section, name, default, required))
        
    def add_attr_path(self, name, default=None, required=False, hidden=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_path(self.section, name, default, required))

